"""
NetVizCorpy - Visualisation tool for Corporate group(s) Network as available on WikiData in Python
Copyright (c) 2025 Zsofia Baruwa

Documentation: https://
Examples: https://
Source: https://github.com/zb15/NetVizCorpy
"""
# Load version info
try:
    from .version import __version__
except ImportError:
    __version__ = "unknown"

# Import key classes for top-level access
from .query import Querier
from .identifyQIDs import Searcher
from .deriveNetwork import NetworkBuilder
from .postProcess import Cleaner
from .visualiseNetwork import Visualiser

# Define public API
__all__ = [
    "Querier",
    "Searcher",
    "NetworkBuilder",
    "Cleaner",
    "Visualiser",
]



from pyvis.network import Network
import seaborn as sns
import colorsys
from collections import Counter


# define a function that takes a degree and a scaling factor and returns a size for the node
def degree_to_size(degree, scaling_factor):
    # multiply degree by scaling factor to get base size
    base_size = degree * scaling_factor
    # if the p and c revenue in the pandas dataframe and it is not NaN, use it to adjust the base size
    # if not np.isnan(revenues):
    #    base_size += revenues/1000
    # add minimum size to base size to ensure no node is too small
    min_size = 3
    final_size = base_size + min_size
    # return final size as an integer
    return float(final_size)


# define a function that converts an rgb tuple to an rgba string with a pastel factor
def rgb_to_rgba(rgb, pastel_factor):
    # convert rgb values to hls values
    h, l, s = colorsys.rgb_to_hls(*rgb)
    # increase lightness by multiplying with pastel factor
    l *= pastel_factor
    # convert hls values back to rgb values
    r, g, b = colorsys.hls_to_rgb(h, l, s)
    # normalize rgb values to be between 0 and 255
    r = int(r * 255)
    g = int(g * 255)
    b = int(b * 255)
    # add alpha value of 0.7
    a = 0.7
    # format rgba values as a string
    rgba = f'rgba({r}, {g}, {b}, {a})'
    return rgba


class Visualiser:
    def __init__(self, input, network_name):
        self.input = input
        self.name = network_name

    def visualise_b2b_network(self):
        final_df = self.input
        # create network object
        net = Network(
            notebook=True,
            directed=True,  # directed graph
            bgcolor="snow",  # background color of graph
            font_color="navy",  # use navy for node labels
            cdn_resources='in_line',  # make sure Jupyter notebook can display correctly
            height="1000px",  # height of chart
            width="100%",  # fill the entire width
            select_menu=True,  # user can choose from campany name list
            filter_menu=True,  # user can search for colors - indicators for industries
            neighborhood_highlight=True,  # clicking on a node highlights its connections and grays out others
        )

        # create lists of nodes and edges from dataframe columns
        nodes = []
        edges = []
        for parent, child, p_industry, c_industry, top_p_industry, top_c_industry, parent_country, \
                parent_continent, child_country, child_continent, p_revenue, p_revenue_date, \
                c_revenue, c_revenue_date, proportion, endtime in zip(final_df['parent'], \
                                                                      final_df['child'], \
                                                                      final_df['p_industries'], \
                                                                      final_df['c_industries'], \
                                                                      final_df['top_p_industries'], \
                                                                      final_df['top_c_industries'], \
                                                                      final_df['parent_country'], \
                                                                      final_df['parent_continent'], \
                                                                      final_df['child_country'], \
                                                                      final_df['child_continent'], \
                                                                      final_df['p_total_revenue_in_millions'], \
                                                                      final_df['p_total_revenue_date'], \
                                                                      final_df['c_total_revenue_in_millions'], \
                                                                      final_df['c_total_revenue_date'], \
                                                                      final_df['p_proportion'], \
                                                                      final_df['endtime']):
            nodes.append((parent, p_industry, top_p_industry, parent_country, parent_continent, p_revenue, p_revenue_date))
            nodes.append((child, c_industry, top_c_industry, child_country, child_continent, c_revenue, c_revenue_date))
            edges.append((parent, child, proportion, endtime))

        # remove duplicate nodes
        # nodes = list(set(nodes))

        # create a set of unique industries
        industries = set(final_df['top_p_industries']).union(set(final_df['top_c_industries']))

        # creat a set of unique revenues and their dates
        # revenues = set(final_df['p_total_revenue_in_millions']).union(set(final_df['c_total_revenue_in_millions']))
        revenues = [p_revenue, c_revenue]
        revenue_dates = [p_revenue_date, c_revenue_date]
        # convert the revenues to a numeric type
        # revenues = pd.to_numeric(revenues)
        # revenue_dates = set(final_df['p_total_revenue_date']).union(set(final_df['c_total_revenue_date']))

        # get the number of unique industries
        n_colors = len(industries)

        # create a dictionary that maps each unique industry to a pastel color
        pastel_colors = sns.color_palette('muted', n_colors)

        group_color_map = {}
        for i, top_industry in enumerate(industries):
            # use the function with pastel_colors[i] and a pastel factor of 1.2
            group_color_map[top_industry] = rgb_to_rgba(pastel_colors[i], 1.2)

        # create a dictionary that maps each node to its degree
        node_degree_map = Counter()
        for edge in edges:
            node_degree_map[edge[0]] += 1
            node_degree_map[edge[1]] += 1

        # create a dictionary that maps each node to its size
        node_size_map = {}
        for node in node_degree_map:
            # use the function with node_degree_map[node] and a scaling factor of 5
            node_size_map[node] = degree_to_size(node_degree_map[node], 5)

        # define a different color for the highlighted nodes
        highlight_color = 'green'
        # node1 = root_companies[0]
        # node2 = root_companies[1]

        # create a set of unique countries and continents
        countries = set(final_df['parent_country']).union(set(final_df['child_country']))
        continents = set(final_df['parent_continent']).union(set(final_df['child_continent']))

        # add nodes and edges to network object
        for node, industry, top_industry, countries, continents, revenues, revenue_dates in nodes:
            color = group_color_map[top_industry]
            # check if the node name is equal to node1 or node2 and use highlight_color if so
            # if node == root_companies:
            #    color = highlight_color
            # else:
            #    color = group_color_map[top_industry]
            # assign shapes for the continents
            if continents == 'North America':
                node_shape = 'triangle'
            elif continents == 'South America':
                node_shape = 'triangleDown'
            elif continents == 'Europe':
                node_shape = 'star'
            elif continents == 'Africa':
                node_shape = 'diamond'
            elif continents == 'Asia':
                node_shape = 'square'
            elif continents == 'Australia':
                node_shape = 'ellipse'
            else:
                node_shape = 'dot'
            size = node_size_map[node]
            net.add_node(node, label=node, group=top_industry, color=color, shape=node_shape, \
                         title=node + ', ' + countries + ' (' + industry + ')' + ' Total revenue: ' + str(
                             revenues) + ' (' + str(revenue_dates) + ')', value=size)

        for e in edges:
            if e[2] > 0 and e[2] < 0.5:
                net.add_edge(e[0], e[1], title=str(e[2]), arrows={"to": True},
                             color='turquoise')  # value=(e[2]/100), dashes=[6,10,1,10]
            elif e[2] >= 0.5 and e[2] <= 1:
                net.add_edge(e[0], e[1], title=str(e[2]), arrows={"to": True},
                             color='violet')  # value=(e[2]/10),  dashes=False
            elif e[2] == 2:
                net.add_edge(e[0], e[1], title="parent", arrows={"to": True}, color='salmon')  # dashes=[5,5]
            elif e[2] == 3:
                net.add_edge(e[0], e[1], title="unknown value", arrows={"to": True}, color='lime')  # dashes=[1,10]
            elif e[2] == 4:
                net.add_edge(e[0], e[1], title="end date: " + str(e[3]), arrows={"to": True}, color='grey')

        net.repulsion(
            node_distance=100,
            central_gravity=0.2,
            spring_length=200,
            spring_strength=0.05,
            damping=0.09,
        )

        net.show_buttons(filter_='physics')
        # net.show_buttons(filter_=['nodes', 'edges', 'physics'])

        html = net.generate_html()
        html_file = self.name + ".html"
        with open(html_file, mode='w', encoding='utf-8') as fp:
            fp.write(html)
            fp.close()

        # show network graph
        #return net.show('B2B_network_Wiki.html')



import pandas as pd
import numpy as np
from query import Querier
import pycountry_convert as pc


def is_human(wikidata_id):
    # Initializing an empty dictionary to store the information
    info = {}
    querier = Querier(wikidata_id, 9)
    response = querier.connect()
    if response.status_code != 204 and response.headers["content-type"].strip().find("json") > -1:
        try:
            # Parsing the response as JSON
            data = response.json()
            # Looping through the results
            for result in data["results"]["bindings"]:
                # Adding the company's QID to the dictionary if exists
                if "QID" in result:
                    info["QID"] = result["QID"]["value"]
                # Adding the company label to the dictionary if exists
                if "humanLabel" in result:
                    info["name"] = result["humanLabel"]["value"]
                # Adding the industry label to the dictionary if exists
                if "humanDescription" in result:
                    info["descr"] = result["humanDescription"]["value"]
            # Returning the dictionary of information
            return info
        except ValueError:
            return info
    else:
        return info


def get_description(wikidata_id):
    # Initializing an empty dictionary to store the information
    info = {}
    querier = Querier(wikidata_id, 0)
    response = querier.connect()
    if response.status_code != 204 and response.headers["content-type"].strip().find("json") > -1:
        try:
            # Parsing the response as JSON
            data = response.json()

            # Initializing an empty dictionary to store the information
            info = {}
            # Looping through the results
            for result in data["results"]["bindings"]:
                # Adding the company's QID to the dictionary if exists
                if "QID" in result:
                    info["QID"] = result["QID"]["value"]
                # Adding the company label to the dictionary if exists
                if "itemLabel" in result:
                    info["name"] = result["itemLabel"]["value"]
                # Adding the industry label to the dictionary if exists
                if "itemDescription" in result:
                    info["descr"] = result["itemDescription"]["value"]
            # Returning the dictionary of information
            return info
        except ValueError:
            return info
    else:
        return info


def clear_each_df(df):
    # convert endtime column to string and filter out rows with valid dates
    if 'endtime' in df.columns:
        df['endtime'] = df['endtime'].astype(str)
        df1 = df[~df['endtime'].str.contains('\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z')]
    new_df = pd.DataFrame(columns=df.columns)
    # create a set of unique QID and obQID pairs
    qid1_qid2_pairs = set(zip(df1.iloc[:, 1], df1.iloc[:, 8]))
    # loop over the unique pairs and select the first row for each latest pointoftime pair
    for qid1, qid2 in qid1_qid2_pairs:
        temp_df = df1.loc[(df1.iloc[:, 1] == qid1) & (df1.iloc[:, 8] == qid2)]
        temp_df = temp_df.drop_duplicates().reset_index(drop=True)
        if len(temp_df) == 1:
            new_row = temp_df.iloc[0]
            new_df = pd.concat([new_df, new_row.to_frame().T])
        if len(temp_df) >= 2:
            # Sort the DataFrame by the pointoftime column in descending order.
            df_filtered = temp_df.sort_values(by='pointoftime', ascending=False)
            # Get the first row of the sorted DataFrame.
            first_row = df_filtered.iloc[0]
            new_df = pd.concat([new_df, first_row.to_frame().T])

    # reset index and drop duplicates
    new_df.reset_index(drop=True, inplace=True)
    new_df = new_df.drop_duplicates().reset_index(drop=True)

    from pandas.errors import SettingWithCopyWarning
    import warnings
    warnings.filterwarnings('ignore', category=SettingWithCopyWarning)
    # add those relations that has an endtime to the new dataframes
    if 'endtime' in df.columns:
        df['endtime'] = df['endtime'].astype(str)
        df2 = df[df['endtime'].str.contains('\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z')]
        df2.loc[:, 'proportion'] = 4
        df2['endtime'] = df2['endtime'].str[:10]
        df2 = df2.drop_duplicates()
    new_df = pd.concat([new_df, df2]).reset_index(drop=True)

    return new_df


def drop_dupl_joined(new_df2):
    ### remove the duplicates from the joined datasets
    # Step 1. save the single parent-child pairs in the new3_df2
    #         if there are 2 or more of the same parent-child pair, then \
    #         drop the ones within these that has blank p_proportion.
    # Step 2. in a new list comprehension (temp_step2_df) save the ones that now became single parent-child.
    #         if there are 2 or more and if one of the p_prop value is == 2, then \
    ##        drop this row with value == 2
    #         Finally, if  there are 2 or more and their p_prop value is equal, then keep only the last
    ##        (this will allow more values to pass through if they differ)
    # STEP 1
    # Use a set to store the unique cQID values
    cQID_set = set(new_df2['cQID'])
    # Use a generator expression to iterate over the pQID values
    pQID_gen = (i for i in new_df2['pQID'])
    # Use a list comprehension to filter the rows based on pQID and cQID values
    temp_new3_df2 = [new_df2.loc[(new_df2['pQID'] == i) & (new_df2['cQID'] == j)]
                     for i in pQID_gen for j in cQID_set]
    # Use another list comprehension to process the filtered rows based on the
    # length and p_proportion values
    new3_df2 = pd.concat([row.iloc[0].to_frame().T if len(row) == 1
                          else row.dropna(subset=['p_proportion']).loc[row['p_proportion'].notnull()]
                          for row in temp_new3_df2])
    new3_df2 = new3_df2.drop_duplicates().reset_index(drop=True)
    # STEP 2
    # Use a set to store the unique cQID values
    cQID_set2 = set(new3_df2['cQID'])
    # Use a generator expression to iterate over the pQID values
    pQID_gen2 = (i for i in new3_df2['pQID'])
    # Use a list comprehension to filter the rows based on pQID and cQID values
    temp_step2_df = [new3_df2.loc[(new3_df2['pQID'] == i) & (new3_df2['cQID'] == j)]
                     for i in pQID_gen2 for j in cQID_set2]
    # Drop the rows with p_proportion value equal to 2 when there is already a record with proportion value
    final_df = pd.concat([row.iloc[0].to_frame().T if len(row) == 1 \
                              else row.loc[row['p_proportion'] != 2] for row in temp_step2_df])
    # Keep only the last row if there are multiple rows with the same pQID and cQID
    # values and equal p_proportion values
    final_df = final_df.drop_duplicates(subset=['pQID', 'cQID', 'p_proportion'], keep='last')
    final_df = final_df.drop_duplicates().reset_index(drop=True)
    # fill the blank p_proportion values with number 3, this will mean unknown on the graph
    final_df['p_proportion'] = final_df['p_proportion'].fillna(3)
    return final_df


def top_industries(final_df):
    # create a dictionary of industry frequencies
    # split industry columns by ";" and create separate columns
    df_split1 = final_df["p_industries"].str.split(";", expand=True)
    df_split2 = final_df["c_industries"].str.split(";", expand=True)
    # join the two dataframes horizontally
    df_split = pd.concat([df_split1, df_split2], axis=1)
    # reshape the dataframe so that each expression is in a single column
    df_melt = df_split.melt(value_name="expression")
    # drop the variable column
    df_melt = df_melt.drop("variable", axis=1)
    # count the frequency of each expression and convert to dictionary
    freq = df_melt["expression"].value_counts().to_dict()

    # define a custom function to pick the expression with highest frequency
    def pick_expression(lst):
        # split the list by semicolon
        lst = lst.split(";")
        # initialize the best expression and frequency
        best_exp = None
        best_freq = 0
        # loop through the list
        for exp in lst:
            # get the frequency of the expression
            freq_exp = freq.get(exp, 0)
            # if the frequency is higher than the best frequency, update the best expression and frequency
            if freq_exp > best_freq:
                best_exp = exp
                best_freq = freq_exp
        # return the best expression
        return best_exp

    # apply the custom function to each cell in col2
    final_df["top_p_industries"] = final_df["p_industries"].apply(pick_expression)
    final_df["top_c_industries"] = final_df["c_industries"].apply(pick_expression)

    # list of columns to be processed
    cols = ["top_p_industries", "top_c_industries", "p_industries", "c_industries"]
    # replace the empty strings with 'unknown' and strip leading spaces
    for col in cols:
      final_df[col] = final_df[col].replace("", np.nan)
      mask = final_df[col].isna()
      final_df.loc[mask, col] = "unknown"
      final_df[col] = final_df[col].str.lstrip()
    
    # make a list of all those parent QIDs where the industry is "unknown"
    unique_values1 = final_df.loc[(final_df['top_p_industries'] == 'unknown'), 'pQID'].drop_duplicates().values.tolist()
    # making sure only those QIDs in unique_values that starts with "Q"
    unique_values1 = [item for item in unique_values1 if item.startswith("Q")]
    # create a dataframe for the results
    p_human_df1 = pd.DataFrame(columns=["QID", "name", "descr"])
    # initialising a new list to store the results
    results = []
    # Loop through the values
    for id in unique_values1:
        # Run the module with the value as input and get the result
        result = pd.DataFrame(is_human(id), index=[0])
        # Appending the new dataframe to the list
        results.append(result)

    # join the new results to the dataframe
    p_human_df1 = pd.concat([p_human_df1] + results, axis=0, join='outer').drop_duplicates().reset_index(drop=True)

    # apply humans and their descriptions to new_df2 dataframe
    for index, row in final_df.iterrows():
        if row['pQID'] in p_human_df1['QID'].values:
            # get the index of the matching row in p_human_df1
            match_index = p_human_df1[p_human_df1['QID'] == row['pQID']].index
            # get the value from the "descr" column of p_human_df1
            descr_value = p_human_df1.loc[match_index[0], 'descr']
            # update the values in new_df2
            final_df.loc[index, 'p_industries'] = descr_value
            final_df.loc[index, 'top_p_industries'] = "human"

    # make a list of all those QIDs where the industry is "unknown"
    unique_values3 = final_df.loc[(final_df['top_p_industries'] == 'unknown'), 'pQID'].drop_duplicates().values.tolist()
    unique_values4 = final_df.loc[(final_df['top_c_industries'] == 'unknown'), 'cQID'].drop_duplicates().values.tolist()
    # making sure only those QIDs in unique_values that starts with "Q"
    unique_values3 = [item for item in unique_values3 if item.startswith("Q")]
    unique_values4 = [item for item in unique_values4 if item.startswith("Q")]
    # create a dataframe for the results
    p_item_df1 = pd.DataFrame(columns=["QID", "name", "descr"])
    # initialising a new list to store the results
    results = []
    # Loop through the values
    for id in unique_values3:
        # Run the module with the value as input and get the result
        result = pd.DataFrame(get_description(id), index=[0])
        # Appending the new dataframe to the list
        results.append(result)

    # join the new results to the dataframe
    p_item_df1 = pd.concat([p_item_df1] + results, axis=0, join='outer').drop_duplicates().reset_index(drop=True)

    # create a dataframe for the results
    c_item_df2 = pd.DataFrame(columns=["QID", "name", "descr"])
    # initialising a new list to store the results
    results2 = []
    # Loop through the values
    for id2 in unique_values4:
        # Run the module with the value as input and get the result
        result2 = pd.DataFrame(get_description(id2), index=[0])
        # Appending the new dataframe to the list
        results2.append(result2)

    # join the new results to the dataframe
    c_item_df2 = pd.concat([c_item_df2] + results2, axis=0, join='outer').drop_duplicates().reset_index(drop=True)

    # apply descriptions for unknown parents to new_df2 dataframe
    for index, row in final_df.iterrows():
        if row['pQID'] in p_item_df1['QID'].values:
            # get the index of the matching row in p_human_df1
            match_index = p_item_df1[p_item_df1['QID'] == row['pQID']].index
            # get the value from the "descr" column of p_human_df1
            descr_value = p_item_df1.loc[match_index[0], 'descr']
            # update the values in new_df2
            final_df.loc[index, 'p_industries'] = descr_value
            final_df.loc[index, 'top_p_industries'] = "other"

    # apply descriptions for unknown children to new_df2 dataframe
    for index, row in final_df.iterrows():
        if row['cQID'] in c_item_df2['QID'].values:
            # get the index of the matching row in p_human_df1
            match_index2 = c_item_df2[c_item_df2['QID'] == row['cQID']].index
            # get the value from the "descr" column of p_human_df1
            descr_value2 = c_item_df2.loc[match_index2[0], 'descr']
            # update the values in new_df2
            final_df.loc[index, 'c_industries'] = descr_value2
            final_df.loc[index, 'top_c_industries'] = "other"

    # list of columns to be processed
    cols2 = ['p_industries', 'top_p_industries', 'c_industries', 'top_c_industries']
    # convert the values in columns to strings
    for col2 in cols2:
      final_df[col2] = final_df[col2].astype(str)
    return final_df

def country_to_continent(country_name):
    import pycountry_convert as pc
    try:
        if country_name.lower() == 'unknown':
            return 'unknown'
        else:
            country_alpha2 = pc.country_name_to_country_alpha2(country_name)
            country_continent_code = pc.country_alpha2_to_continent_code(country_alpha2)
            country_continent_name = pc.convert_continent_code_to_continent_name(country_continent_code)
            return country_continent_name
    except KeyError:
        return 'unknown'


    # class PostProcessor:
    #    def __init__(self, input):
    #        self.input = input

    # clean each dataset and then join them


class Cleaner:
    def __init__(self, input):
        self.input = input

    def clean_join(self):
        results = self.input
        p_df = results['p_df'][-1]
        ob_df = results['ob_df'][-1]
        s_df = results['s_df'][-1]
        oo_df = results['oo_df'][-1]

        # global final_df, new_p_df, new_ob_df, new_s_df, new_oo_df
        # convert the data columns to strings
        p_df.loc[:, ['pointoftime', 'starttime']] = p_df.loc[:, ['pointoftime', 'starttime']].astype(str)
        ob_df.loc[:, ['pointoftime', 'starttime']] = ob_df.loc[:, ['pointoftime', 'starttime']].astype(str)
        s_df.loc[:, ['pointoftime', 'starttime']] = s_df.loc[:, ['pointoftime', 'starttime']].astype(str)
        oo_df.loc[:, ['pointoftime', 'starttime']] = oo_df.loc[:, ['pointoftime', 'starttime']].astype(str)
        ### remove the duplicates within p dataset - keep the last
        new_p_df = pd.DataFrame(columns=['item', 'QID', 'itemLabel', 'itemcountryLabel', 'industries', \
                                         'revenue', 'revenueDate', \
                                         'parent', 'pQID', 'parentLabel', 'parentcountryLabel', 'pindustries', \
                                         'prevenue', 'prevenueDate', \
                                         'proportion', 'proportionofLabel', 'pointoftime', 'starttime', 'endtime'])
        new_p_df = pd.concat([new_p_df, clear_each_df(p_df)]).reset_index(drop=True)

        ### remove the duplicates within ob dataset - keep the last
        new_ob_df = pd.DataFrame(columns=['item', 'QID', 'itemLabel', 'itemcountryLabel', 'industries', \
                                          'revenue', 'revenueDate', \
                                          'ownedby', 'obQID', 'ownedbyLabel', 'ownedbycountryLabel', 'obindustries', \
                                          'obrevenue', 'obrevenueDate', \
                                          'proportion', 'proportionofLabel', 'pointoftime', 'starttime', 'endtime'])
        new_ob_df = pd.concat([new_ob_df, clear_each_df(ob_df)]).reset_index(drop=True)

        ### remove the duplicates within s dataset - keep the last
        new_s_df = pd.DataFrame(columns=['item', 'QID', 'itemLabel', 'itemcountryLabel', 'industries', \
                                         'revenue', 'revenueDate', \
                                         'subsidiary', 'sQID', 'subsidiaryLabel', 'subsidiarycountryLabel',
                                         'sindustries', \
                                         'srevenue', 'srevenueDate', \
                                         'proportion', 'proportionofLabel', 'pointoftime', 'starttime', 'endtime'])
        new_s_df = pd.concat([new_s_df, clear_each_df(s_df)]).reset_index(drop=True)

        ### remove the duplicates within oo dataset - keep the last
        new_oo_df = pd.DataFrame(columns=['item', 'QID', 'itemLabel', 'itemcountryLabel', 'industries', \
                                          'revenue', 'revenueDate', \
                                          'ownerof', 'ooQID', 'ownerofLabel', 'ownerofcountryLabel', 'ooindustries', \
                                          'oorevenue', 'oorevenueDate', \
                                          'proportion', 'proportionofLabel', 'pointoftime', 'starttime', 'endtime'])
        new_oo_df = pd.concat([new_oo_df, clear_each_df(oo_df)]).reset_index(drop=True)

        ### join together the dataframes into a new_df
        new_df = pd.DataFrame(
            columns=["pQID", "parent", "parent_country", "p_industries", "p_total_revenue", "p_total_revenue_date", \
                     "cQID", "child", "child_country", "c_industries", "c_total_revenue", "c_total_revenue_date", \
                     "p_proportion", "proportionofLabel", \
                     "pointoftime", "starttime", "endtime"])
        # Create an empty list to store the new rows
        new_p_rows = []
        # Loop over the new_p_df dataframe
        for i in range(len(new_p_df)):
            # Create a new row as a dictionary
            new_p_row = {'pQID': new_p_df['pQID'][i], 'parent': new_p_df['parentLabel'][i],
                         'parent_country': new_p_df['parentcountryLabel'][i], \
                         'p_industries': new_p_df['pindustries'][i], 'p_total_revenue': new_p_df['prevenue'][i], \
                         'p_total_revenue_date': new_p_df['prevenueDate'][i], \
                         'cQID': new_p_df['QID'][i], 'child': new_p_df['itemLabel'][i],
                         'child_country': new_p_df['itemcountryLabel'][i], \
                         'c_industries': new_p_df['industries'][i], 'c_total_revenue': new_p_df['revenue'][i], \
                         'c_total_revenue_date': new_p_df['revenueDate'][i], \
                         'p_proportion': new_p_df['proportion'][i],
                         'proportionofLabel': new_p_df['proportionofLabel'][i], \
                         'pointoftime': new_p_df['pointoftime'][i], 'starttime': new_p_df['starttime'][i],
                         'endtime': new_p_df['endtime'][i]}
            # Convert the dictionary to a dataframe and append it to the list
            new_p_rows.append(pd.DataFrame(new_p_row, index=[0]))

        # Concatenate the list of dataframes with the original dataframe
        new_df = pd.concat([new_df] + new_p_rows, ignore_index=True)
        new_df['p_proportion'] = new_df['p_proportion'].fillna(2)
        # Create an empty list to store the new rows
        new_ob_rows = []
        for i in range(len(new_ob_df)):
            new_ob_row = {'pQID': new_ob_df['obQID'][i], 'parent': new_ob_df['ownedbyLabel'][i],
                          'parent_country': new_ob_df['ownedbycountryLabel'][i], \
                          'p_industries': new_ob_df['obindustries'][i], 'p_total_revenue': new_ob_df['obrevenue'][i], \
                          'p_total_revenue_date': new_ob_df['obrevenueDate'][i], \
                          'cQID': new_ob_df['QID'][i], 'child': new_ob_df['itemLabel'][i],
                          'child_country': new_ob_df['itemcountryLabel'][i], \
                          'c_industries': new_ob_df['industries'][i], 'c_total_revenue': new_ob_df['revenue'][i], \
                          'c_total_revenue_date': new_ob_df['revenueDate'][i], \
                          'p_proportion': new_ob_df['proportion'][i],
                          'proportionofLabel': new_ob_df['proportionofLabel'][i], \
                          'pointoftime': new_ob_df['pointoftime'][i], 'starttime': new_ob_df['starttime'][i],
                          'endtime': new_ob_df['endtime'][i]}
            new_ob_rows.append(pd.DataFrame(new_ob_row, index=[0]))

        # Concatenate the list of dataframes with the original dataframe
        new_df = pd.concat([new_df] + new_ob_rows, ignore_index=True)
        # Create an empty list to store the new rows
        new_s_rows = []
        for i in range(len(new_s_df)):
            new_s_row = {'pQID': new_s_df['QID'][i], 'parent': new_s_df['itemLabel'][i],
                         'parent_country': new_s_df['itemcountryLabel'][i], \
                         'p_industries': new_s_df['industries'][i], 'p_total_revenue': new_s_df['revenue'][i], \
                         'p_total_revenue_date': new_s_df['revenueDate'][i], \
                         'cQID': new_s_df['sQID'][i], 'child': new_s_df['subsidiaryLabel'][i],
                         'child_country': new_s_df['subsidiarycountryLabel'][i], \
                         'c_industries': new_s_df['sindustries'][i], 'c_total_revenue': new_s_df['srevenue'][i], \
                         'c_total_revenue_date': new_s_df['srevenueDate'][i], \
                         'p_proportion': new_s_df['proportion'][i],
                         'proportionofLabel': new_s_df['proportionofLabel'][i], \
                         'pointoftime': new_s_df['pointoftime'][i], 'starttime': new_s_df['starttime'][i],
                         'endtime': new_s_df['endtime'][i]}
            new_s_rows.append(pd.DataFrame(new_s_row, index=[0]))
        # Concatenate the list of dataframes with the original dataframe
        new_df = pd.concat([new_df] + new_s_rows, ignore_index=True)
        # Create an empty list to store the new rows
        new_oo_rows = []
        for i in range(len(new_oo_df)):
            new_oo_row = {'pQID': new_oo_df['QID'][i], 'parent': new_oo_df['itemLabel'][i],
                          'parent_country': new_oo_df['itemcountryLabel'][i], \
                          'p_industries': new_oo_df['industries'][i], 'p_total_revenue': new_oo_df['revenue'][i], \
                          'p_total_revenue_date': new_oo_df['revenueDate'][i], \
                          'cQID': new_oo_df['ooQID'][i], 'child': new_oo_df['ownerofLabel'][i],
                          'child_country': new_oo_df['ownerofcountryLabel'][i], \
                          'c_industries': new_oo_df['ooindustries'][i], 'c_total_revenue': new_oo_df['oorevenue'][i], \
                          'c_total_revenue_date': new_oo_df['oorevenueDate'][i], \
                          'p_proportion': new_oo_df['proportion'][i],
                          'proportionofLabel': new_oo_df['proportionofLabel'][i], \
                          'pointoftime': new_oo_df['pointoftime'][i], 'starttime': new_oo_df['starttime'][i],
                          'endtime': new_oo_df['endtime'][i]}
            new_oo_rows.append(pd.DataFrame(new_oo_row, index=[0]))
        # Concatenate the list of dataframes with the original dataframe
        new_df = pd.concat([new_df] + new_oo_rows, ignore_index=True)
        new_df2 = new_df.drop_duplicates().reset_index(drop=True)

        # drop the duplicates from the joined dataframe with various rules defined by drop_dupl_joined() function
        final_df = drop_dupl_joined(new_df2)
        # clear up industries and find the top ones for classification. unknown ones: get description (also checks if its human)
        final_df = top_industries(final_df)

        # replace unknow country names with "unknown"
        final_df[["parent_country", "child_country"]] = final_df[["parent_country", "child_country"]].replace({"": np.nan, np.nan: "unknown"})

        # apply the country_to_continent function to get the continents the companies are located
        final_df['parent_continent'] = final_df['parent_country'].apply(country_to_continent)
        final_df['child_continent'] = final_df['child_country'].apply(country_to_continent)

        # trim the date for revenue
        final_df['p_total_revenue_date'] = final_df['p_total_revenue_date'].astype(str).str[:4]
        final_df['c_total_revenue_date'] = final_df['c_total_revenue_date'].astype(str).str[:4]

        # divide revenue by 1M and rename colum names
        final_df.loc[final_df['p_total_revenue'].notnull(), 'p_total_revenue'] = final_df.loc[final_df[
            'p_total_revenue'].notnull(), 'p_total_revenue'].astype('float') / 1000000
        final_df.loc[final_df['c_total_revenue'].notnull(), 'c_total_revenue'] = final_df.loc[final_df[
            'c_total_revenue'].notnull(), 'c_total_revenue'].astype('float') / 1000000
        final_df = final_df.rename(
            columns={'p_total_revenue': 'p_total_revenue_in_millions',
                     'c_total_revenue': 'c_total_revenue_in_millions'})
        # format the revenue values and convert it to string
        final_df.loc[final_df['p_total_revenue_in_millions'].notnull(), 'p_total_revenue_in_millions'] = final_df.loc[
            final_df['p_total_revenue_in_millions'].notnull(), 'p_total_revenue_in_millions'].apply(lambda x: f"{x:,}M")
        final_df.loc[final_df['c_total_revenue_in_millions'].notnull(), 'c_total_revenue_in_millions'] = final_df.loc[
            final_df['c_total_revenue_in_millions'].notnull(), 'c_total_revenue_in_millions'].apply(lambda x: f"{x:,}M")
        #fill the empty revenue values with "no data"
        final_df['p_total_revenue_in_millions'] = final_df['p_total_revenue_in_millions'].fillna("no data")
        final_df['c_total_revenue_in_millions'] = final_df['c_total_revenue_in_millions'].fillna("no data")
        return final_df



import pandas as pd
import numpy as np
from query import Querier


# Checking if entity has a parent, and if so how many
def has_parent(wikidata_id):
    querier = Querier(wikidata_id, 1)
    response = querier.connect()
    if response.status_code != 204 and response.headers["content-type"].strip().find("json") > -1:
        try:
            # Parsing the response as JSON
            data = response.json()
            # Extracting all the results
            results = data["results"]["bindings"]
            # Check if results is empty
            if not results:
                # Return zero
                return 0
            else:
                # Get the count value
                counts = results[0]["count"]["value"]
                # Make counts integers
                counts = int(counts)
                # Return the count as an integer
                return counts
        except ValueError:
            return 0
    else:
        return 0


def get_company_parent_info(wikidata_id):
    querier = Querier(wikidata_id, 2)
    response = querier.connect()
    if response.status_code != 204 and response.headers["content-type"].strip().find("json") > -1:
        try:
            # extracts the results and bindings keys from the JSON response returned by the Wikidata SPARQL endpoint
            data = response.json()['results']['bindings']
            # converts the JSON data into a pandas DataFrame
            df = pd.json_normalize(data)
            #  checks if an element is a dictionary and if so, extracts the value key from it
            df = df.applymap(lambda x: x['value'] if isinstance(x, dict) else x)
            # selects only the columns that end with .value.
            df = df[[col for col in df.columns if col.endswith('.value')]]
            # removes the .value suffix from each column name
            df.columns = [col.split('.')[0] for col in df.columns]
            return df
        except ValueError:
            return pd.DataFrame()
    else:
        return pd.DataFrame()


def has_ownedby(wikidata_id):
    querier = Querier(wikidata_id, 3)
    response = querier.connect()
    if response.status_code != 204 and response.headers["content-type"].strip().find("json") > -1:
        try:
            # Parsing the response as JSON
            data = response.json()
            # Extracting all the results
            results = data["results"]["bindings"]
            # Check if results is empty
            if not results:
                # Return zero
                return 0
            else:
                # Get the count value
                counts = results[0]["count"]["value"]
                # Make counts integers
                counts = int(counts)
                # Return the count as an integer
                return counts
        except ValueError:
            return 0
    else:
        return 0


# Defining a function that takes a Wikidata ID and returns a dictionary of information about its industry, parent organizations and subsidiaries
def get_company_ownedby_info(wikidata_id):
    querier = Querier(wikidata_id, 4)
    response = querier.connect()
    if response.status_code != 204 and response.headers["content-type"].strip().find("json") > -1:
        try:
            # extracts the results and bindings keys from the JSON response returned by the Wikidata SPARQL endpoint
            data = response.json()['results']['bindings']
            # converts the JSON data into a pandas DataFrame
            df = pd.json_normalize(data)
            #  checks if an element is a dictionary and if so, extracts the value key from it
            df = df.applymap(lambda x: x['value'] if isinstance(x, dict) else x)
            # selects only the columns that end with .value.
            df = df[[col for col in df.columns if col.endswith('.value')]]
            # removes the .value suffix from each column name
            df.columns = [col.split('.')[0] for col in df.columns]
            return df
        except ValueError:
            return pd.DataFrame()
    else:
        return pd.DataFrame()


# Checking if entity has a subsidiary (which is also a type of company), and if so how many
def has_subsidiary(wikidata_id):
    querier = Querier(wikidata_id, 5)
    response = querier.connect()
    if response.status_code != 204 and response.headers["content-type"].strip().find("json") > -1:
        try:
            # Parsing the response as JSON
            data = response.json()
            # Extracting all the results
            results = data["results"]["bindings"]
            # Check if results is empty
            if not results:
                # Return zero
                return 0
            else:
                # Get the count value
                counts = results[0]["count"]["value"]
                # Make counts integers
                counts = int(counts)
                # Return the count as an integer
                return counts
        except ValueError:
            return 0
    else:
        return 0


# getting company's subsidaries
def get_company_subsidiary_info(wikidata_id):
    querier = Querier(wikidata_id, 6)
    response = querier.connect()
    if response.status_code != 204 and response.headers["content-type"].strip().find("json") > -1:
        try:
            # extracts the results and bindings keys from the JSON response returned by the Wikidata SPARQL endpoint
            data = response.json()['results']['bindings']
            # converts the JSON data into a pandas DataFrame
            df = pd.json_normalize(data)
            #  checks if an element is a dictionary and if so, extracts the value key from it
            df = df.applymap(lambda x: x['value'] if isinstance(x, dict) else x)
            # selects only the columns that end with .value.
            df = df[[col for col in df.columns if col.endswith('.value')]]
            # removes the .value suffix from each column name
            df.columns = [col.split('.')[0] for col in df.columns]
            return df
        except ValueError:
            return pd.DataFrame()
    else:
        return pd.DataFrame()
    # return df


# Checking if entity has a ownerof (which is also a type of company), and if so how many
def has_ownerof(wikidata_id):
    querier = Querier(wikidata_id, 7)
    response = querier.connect()
    if response.status_code != 204 and response.headers["content-type"].strip().find("json") > -1:
        try:
            # Parsing the response as JSON
            data = response.json()
            # Extracting all the results
            results = data["results"]["bindings"]
            # Check if results is empty
            if not results:
                # Return zero
                return 0
            else:
                # Get the count value
                counts = results[0]["count"]["value"]
                # Make counts integers
                counts = int(counts)
                # Return the count as an integer
                return counts
        except ValueError:
            return 0


# Defining a function that takes a Wikidata ID and returns a dictionary of information about its industry, parent organizations and subsidiaries
def get_company_ownerof_info(wikidata_id):
    querier = Querier(wikidata_id, 8)
    response = querier.connect()
    if response.status_code != 204 and response.headers["content-type"].strip().find("json") > -1:
        try:
            # extracts the results and bindings keys from the JSON response returned by the Wikidata SPARQL endpoint
            data = response.json()['results']['bindings']
            # converts the JSON data into a pandas DataFrame
            df = pd.json_normalize(data)
            #  checks if an element is a dictionary and if so, extracts the value key from it
            df = df.applymap(lambda x: x['value'] if isinstance(x, dict) else x)
            # selects only the columns that end with .value.
            df = df[[col for col in df.columns if col.endswith('.value')]]
            # removes the .value suffix from each column name
            df.columns = [col.split('.')[0] for col in df.columns]
            return df
        except ValueError:
            return pd.DataFrame()
    else:
        return pd.DataFrame()


def tidy_each_df(df):
    # split companies' industries string and return them as list, then drop duplicates, and finally return as string again
    df.iloc[:, 4] = df.iloc[:, 4].str.split("; ").apply(lambda x: "; ".join(set(word.lower() for word in x)))
    # drop parents' industries duplicates, then return as string again
    df.iloc[:, 11] = df.iloc[:, 11].str.split("; ").apply(lambda x: "; ".join(set(word.lower() for word in x)))
    # check if all proprtions are between 0 and 1, if above 1, then devide by 100, otherwise leave it
    if 'proportion' in df.columns:
        df['proportion'] = df['proportion'].astype(float)
        df['proportion'] = np.where(df['proportion'].isnull(), df['proportion'],
                                    np.where(df['proportion'] > 1, df['proportion'] / 100,
                                             df['proportion']))
    df = df.drop_duplicates().reset_index(drop=True)
    return df


class NetworkBuilder:
    def __init__(self, qids, levels):
        self.qids=qids
        self.levels=levels

    def get_companies_network(self):
        # Create an empty dataframe for parent
        p_df = pd.DataFrame(columns=['item', 'QID', 'itemLabel', 'itemcountryLabel', 'industries', \
                                     'revenue', 'revenueDate', \
                                     'parent', 'pQID', 'parentLabel', 'parentcountryLabel', 'pindustries', \
                                     'prevenue', 'prevenueDate', \
                                     'proportion', 'proportionofLabel', 'pointoftime', 'starttime', 'endtime'])
        # Create an empty dataframe for owned by
        ob_df = pd.DataFrame(columns=['item', 'QID', 'itemLabel', 'itemcountryLabel', 'industries', \
                                      'revenue', 'revenueDate', \
                                      'ownedby', 'obQID', 'ownedbyLabel', 'ownedbycountryLabel', 'obindustries', \
                                      'obrevenue', 'obrevenueDate', \
                                      'proportion', 'proportionofLabel', 'pointoftime', 'starttime', 'endtime'])
        # Create an empty dataframe for subsidiary
        s_df = pd.DataFrame(columns=['item', 'QID', 'itemLabel', 'itemcountryLabel', 'industries', \
                                     'revenue', 'revenueDate', \
                                     'subsidiary', 'sQID', 'subsidiaryLabel', 'subsidiarycountryLabel', 'sindustries', \
                                     'srevenue', 'srevenueDate', \
                                     'proportion', 'proportionofLabel', 'pointoftime', 'starttime', 'endtime'])
        # Create an empty dataframe for owner of
        oo_df = pd.DataFrame(columns=['item', 'QID', 'itemLabel', 'itemcountryLabel', 'industries', \
                                      'revenue', 'revenueDate', \
                                      'ownerof', 'ooQID', 'ownerofLabel', 'ownerofcountryLabel', 'ooindustries', \
                                      'oorevenue', 'oorevenueDate', \
                                      'proportion', 'proportionofLabel', 'pointoftime', 'starttime', 'endtime'])

        # initialize a list to store the results of each run
        results = {"p_df": [], "ob_df": [], "s_df": [], "oo_df": []}
        # making sure only those QIDs in root_QIDs that starts with "Q"
        root_QIDs = [item for item in self.qids if item.startswith("Q")]
        # Find the maximum number of levels
        levels = self.levels
        max_levels = max(levels)
        # Iterate from 0 to max_levels-1
        for j in range(max_levels):
            # Iterate over each element of the levels tuple
            for i in range(len(levels)):
                # Check if j is less than the value of levels[i]
                if j < levels[i]:
                    # Iterate over each element of the item list
                    if i == 0:
                        # making sure only those QIDs in root_QIDs that starts with "Q"
                        root_QIDs = [item for item in root_QIDs if item.startswith("Q")]
                        ## HAS PARENT
                        # make a new parent QID list with only those that has parent
                        new_p_QIDs = []
                        for id in root_QIDs:
                            parent_count = has_parent(id)
                            if parent_count > 0:
                                new_p_QIDs.append(id)
                        ## PARENT INFO
                        # Check if new_p_QIDs is empty
                        # get the parent info from the new QID list and concat it with the p_df dataframe
                        for id in new_p_QIDs:
                            p_df = pd.concat([p_df, get_company_parent_info(id)], ignore_index=True)
                        # tidy up industries and proportions with function tidy_each_df()
                        p_df = tidy_each_df(p_df)
                    elif i == 1:
                        # making sure only those QIDs in root_QIDs that starts with "Q"
                        root_QIDs = [item for item in root_QIDs if item.startswith("Q")]
                        ## HAS OWNED BY
                        # make a new ownedby QID list with only those companies that has ownedby
                        new_ob_QIDs = []
                        for id in root_QIDs:
                            ownedby_count = has_ownedby(id)
                            if ownedby_count > 0:
                                new_ob_QIDs.append(id)
                        ## GET OWNED BY INFO
                        for id in new_ob_QIDs:
                            ob_df = pd.concat([ob_df, get_company_ownedby_info(id)], ignore_index=True)
                        # tidy up industries and proportions with function tidy_each_df()
                        ob_df = tidy_each_df(ob_df)
                    elif i == 2:
                        # making sure only those QIDs in root_QIDs that starts with "Q"
                        root_QIDs = [item for item in root_QIDs if item.startswith("Q")]
                        ## HAS SUBSIDIARY
                        # make a new subsidiary QID list with only those companies that has subsidiary
                        new_s_QIDs = []
                        for id in root_QIDs:
                            subsidiary_count = has_subsidiary(id)
                            if subsidiary_count > 0:
                                new_s_QIDs.append(id)
                        ## GET SUBSIDIARY INFO
                        # get the subsidiaries info from the new QID list and concat it with the s_df dataframe
                        for id in new_s_QIDs:
                            s_df = pd.concat([s_df, get_company_subsidiary_info(id)], ignore_index=True)
                        # tidy up industries and proportions with function tidy_each_df()
                        s_df = tidy_each_df(s_df)
                    elif i == 3:
                        # making sure only those QIDs in root_QIDs that starts with "Q"
                        root_QIDs = [item for item in root_QIDs if item.startswith("Q")]
                        ## HAS OWNER OF
                        # make a new ownerof QID list with only those companies that has owner of
                        new_oo_QIDs = []
                        for id in root_QIDs:
                            ownerof_count = has_ownerof(id)
                            if ownerof_count > 0:
                                new_oo_QIDs.append(id)
                        ## GET OWNER OF INFO
                        # get the owner of info from the new QID list and concat it with the oo_df dataframe
                        for id in new_oo_QIDs:
                            oo_df = pd.concat([oo_df, get_company_ownerof_info(id)], ignore_index=True)
                        # tidy up industries and proportions with function tidy_each_df()
                        oo_df = tidy_each_df(oo_df)

            # append results after each number of runs
            results['p_df'].append(p_df)
            results['ob_df'].append(ob_df)
            results['s_df'].append(s_df)
            results['oo_df'].append(oo_df)
            ## JOIN NEW UNIQUE QIDS
            # getting all the QIDs that will needs to be queried again
            # Store dataframes in a list
            df_list = [p_df, ob_df, s_df, oo_df]
            # Store column names in a list
            col_list = ['pQID', 'obQID', 'sQID', 'ooQID']
            # Create an empty list to store the column values
            val_list = []
            # Loop over the dataframes and the column names and append the column values to the list
            for df, col in zip(df_list, col_list):
                val_list.append(df[col].to_list())
            # collect the unique values from val_list
            new_list = []
            [new_list.append(x) for x in val_list if x not in new_list]
            # Convert the list to a pandas series
            original_series = pd.Series(new_list)
            # Use the explode method to split the nested lists into separate rows
            exploded_series = original_series.explode()
            # Use the drop_duplicates method to remove any duplicate values
            unique_series = exploded_series.drop_duplicates()
            # Convert the series back to a list
            new_QIDs = unique_series.tolist()
            # clean new_QIDs from 'nan' string values
            new_QIDs = [x for x in new_QIDs if str(x) != 'nan']
            # get a list of the new unique QIDs
            new1_QIDs = list(set(new_QIDs) - set(root_QIDs))
            # keep only those QIDs that starts with "Q"
            new2_QIDs = [item for item in new1_QIDs if item.startswith("Q")]
            # rename the new list of QIDs as root_QIDs
            root_QIDs = new2_QIDs

        # return the list of results
        return results
